"""CSV Pipeline tbox pytest"""

import json
import os

import pytest

test_folder = os.path.dirname(os.path.abspath(__file__))
working_folder = os.path.join(test_folder, "input")
output_folder = os.path.join(test_folder, "output")

mapping_folder = os.path.join(working_folder, "mapping")
raw_data = os.path.join(working_folder, "data", "classes.csv")
expected = os.path.join(output_folder, "output_csv_parser.ttl")


@pytest.mark.parametrize("extension", ["xlsx", "json", "csv", dict])
def test_csv_pipeline_tbox(extension) -> None:
    from rdflib import Graph

    from data2rdf.warnings import MappingMissmatchWarning

    from data2rdf import Data2RDF, Parser  # isort:skip

    if isinstance(extension, str):
        mapping = os.path.join(mapping_folder, f"mapping.{extension}")

    else:
        path = os.path.join(mapping_folder, "mapping.json")
        with open(path, encoding="utf-8") as file:
            mapping = json.load(file)

    with pytest.warns(
        MappingMissmatchWarning, match="Data for key"
    ) as warnings:
        pipeline = Data2RDF(
            mode="tbox",
            raw_data=raw_data,
            mapping=mapping,
            parser=Parser.csv,
            parser_args={
                "column_sep": ";",
                "suffix_location": "Ontological concept ID",
                "ontology_title": "Test Ontology",
                "authors": ["Jane Doe"],
                "version_info": "1.0.0",
            },
            config={
                "base_iri": "https://w3id.org/dimat",
            },
        )

    missmatches = [
        warning
        for warning in warnings
        if warning.category == MappingMissmatchWarning
    ]
    assert len(missmatches) == 1

    expected_graph = Graph()
    expected_graph.parse(expected)

    assert pipeline.graph.isomorphic(expected_graph)


@pytest.mark.parametrize("input_kind", ["path", "content"])
def test_csv_pipeline_inputs_tbox(input_kind) -> None:
    from rdflib import Graph

    from data2rdf.warnings import MappingMissmatchWarning

    from data2rdf import (  # isort:skip
        Data2RDF,
        Parser,
    )

    if input_kind == "path":
        input_obj = raw_data
    elif input_kind == "content":
        with open(raw_data, encoding="utf-8") as file:
            input_obj = file.read()

    with pytest.warns(
        MappingMissmatchWarning, match="Data for key"
    ) as warnings:
        pipeline = Data2RDF(
            mode="tbox",
            raw_data=input_obj,
            mapping=os.path.join(mapping_folder, "mapping.json"),
            parser=Parser.csv,
            parser_args={
                "column_sep": ";",
                "suffix_location": "Ontological concept ID",
                "ontology_title": "Test Ontology",
                "authors": ["Jane Doe"],
                "version_info": "1.0.0",
            },
            config={
                "base_iri": "https://w3id.org/dimat",
            },
        )

    missmatches = [
        warning
        for warning in warnings
        if warning.category == MappingMissmatchWarning
    ]
    assert len(missmatches) == 1

    expected_graph = Graph()
    expected_graph.parse(expected)

    assert pipeline.graph.isomorphic(expected_graph)



"""JSON pipeline pytest"""


def test_pipeline_dict_custom_properties() -> None:
    from rdflib import Graph

    from data2rdf import Data2RDF, Parser

    data = {
        "data": {
            "name": "Jane Doe",
            "measurement": "Continuous Stiffness Measurement",
        }
    }

    mapping = [
        {
            "value_location": "data.name",
            "value_relation": "http://xmlns.com/foaf/0.1/name",
            "iri": "https://w3id.org/emmo/domain/characterisation-methodology/chameo#Operator",
            "suffix": "Operator1",
        },
        {
            "value_location": "data.measurement",
            "iri": "https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#EMMO_5ca6e1c1-93e9-5e1a-881b-2c2bd38074b1",
            "suffix": "CSM1",
        },
    ]

    addtional_triples = """
    @prefix : <http://abox-namespace-placeholder.org/> .
    @prefix chameo: <https://w3id.org/emmo/domain/characterisation-methodology/chameo#> .

    :CSM1 chameo:hasOperator :Operator1 .
    """

    base_iri = (
        "https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation"
    )

    expected = """
    @prefix foaf: <http://xmlns.com/foaf/0.1/> .
    @prefix ns1: <https://w3id.org/emmo/domain/characterisation-methodology/chameo#> .
    @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
    @prefix ns2: <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#> .
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

    ns2:CSM1 a ns2:EMMO_5ca6e1c1-93e9-5e1a-881b-2c2bd38074b1 ;
             rdfs:label "Continuous Stiffness Measurement"^^xsd:string ;
             ns1:hasOperator ns2:Operator1 .

    ns2:Operator1 a ns1:Operator ;
                  foaf:name "Jane Doe"^^xsd:string .
    """

    pipeline = Data2RDF(
        raw_data=data,
        mapping=mapping,
        parser=Parser.json,
        additional_triples=addtional_triples,
        config={
            "base_iri": base_iri,
            "separator": "#",
            "prefix_name": "nanoindentation",
            "suppress_file_description": True,
        },
    )

    expected_graph = Graph()
    expected_graph.parse(data=expected)

    assert pipeline.graph.isomorphic(expected_graph)



# The csv2rdf conversion requires a set of relations and classes to use for
# the annotation these are defined separately, so that they can be easily
# adapted.
from typing import List, Optional, Union

from pydantic import AnyUrl, ConfigDict, Field, model_validator
from pydantic_settings import BaseSettings


class Config(BaseSettings):
    """Data2RDF configuration"""

    qudt_units: Union[str, AnyUrl] = Field(
        "http://qudt.org/2.1/vocab/unit",
        description="URI to QUDT Unit ontology for unit conversion",
    )

    qudt_quantity_kinds: Union[str, AnyUrl] = Field(
        "http://qudt.org/vocab/quantitykind/",
        description="URI to QUDT quantity kind ontology for unit conversion",
    )

    language: str = Field("en", description="Language for the unit labels")

    base_iri: Union[str, AnyUrl] = Field(
        "https://www.example.org", description="Base IRI for individuals."
    )

    prefix_name: str = Field(
        "fileid",
        description="Prefix used referencing the `base_iri` in the context of the graph.",
    )

    separator: str = Field(
        "/", description="Separator between base IRI and suffix."
    )

    encoding: str = Field("utf-8", description="Encoding used while parsing.")

    data_download_uri: Union[str, AnyUrl] = Field(
        "https://www.example.org/download",
        description="General base iri for downloading the dataframe after uploading",
    )

    graph_identifier: Optional[Union[str, AnyUrl]] = Field(
        None, description="Identifier of the graph to be produced."
    )

    namespace_placeholder: Union[str, AnyUrl] = Field(
        "http://abox-namespace-placeholder.org/",
        description="Placeholder of the extra triples to be replaced with the `base_iri` during the pipeline run.",
    )

    remove_from_unit: List[str] = Field(
        ["[", "]", '"', " "],
        description="Characters which should be removed from the input value for the unit",
    )

    mapping_csv_separator: str = Field(
        ";",
        description="When the mapping file is a csv, the separator to be used for parsing",
    )

    remove_from_datafile: List[str] = Field(
        ['"', "\r", "\n"],
        description="""In plain text parsers, e.g. the CSV-parser,
        there might be the need to remove certain characters when parsing""",
    )

    suppress_file_description: bool = Field(
        False,
        description="""In ABox mode, the pipeline is producing an additional
        subgraph graph for describing the data file in its structure, mime type, etc.
        This will be suppressed if enabled.""",
    )

    exclude_ontology_title: bool = Field(
        False,
        description="In TBox mode, exclude the title of the ontology in the graph.",
    )

    model_config = ConfigDict(extra="ignore")

    @model_validator(mode="after")
    @classmethod
    def validate_config(cls, self: "Config") -> "Config":
        for key, value in self.model_fields.items():
            if isinstance(value, AnyUrl):
                setattr(self, key, str(value))
        return self



"""Data2rdf excel parser"""

import warnings
from io import BytesIO
from typing import Any, Dict, List, Union
from urllib.parse import quote, urljoin

import pandas as pd
from openpyxl import load_workbook
from pydantic import Field

from data2rdf.models.graph import PropertyGraph, QuantityGraph
from data2rdf.utils import make_prefix
from data2rdf.warnings import MappingMissmatchWarning

from .base import ABoxBaseParser, BaseFileParser, TBoxBaseParser

from .utils import (  # isort:skip
    _make_tbox_classes,
    _make_tbox_json_ld,
    _strip_unit,
    _value_exists,
)

from data2rdf.models.mapping import (  # isort:skip
    ABoxExcelMapping,
    TBoxBaseMapping,
    CustomRelationPropertySubgraph,
    CustomRelationQuantitySubgraph,
)


def _load_data_file(
    self: "Union[ExcelTBoxParser, ExcelABoxParser]",
) -> BytesIO:
    """Load excel file as bytes"""
    if isinstance(self.raw_data, str):
        with open(self.raw_data, mode="rb") as file:
            content = BytesIO(file.read())
    elif isinstance(self.raw_data, bytes):
        content = BytesIO(self.raw_data)
    else:
        raise TypeError(
            f"`raw_data` must be of type `str` or `btyes`, not `{type(self.raw_data)}`"
        )
    return content


class ExcelTBoxParser(TBoxBaseParser):
    """
    Parses a data file of type excel in b box mode
    """

    sheet: str = Field(..., description="Name of the sheet for the mapping.")

    header_length: int = Field(
        1, description="Length of the header of the excel sheet", ge=1
    )

    # OVERRIDE
    mapping: Union[str, List[TBoxBaseMapping]] = Field(
        ...,
        description="""File path to the mapping file to be parsed or
        a list with the mapping.""",
    )

    # OVERRIDE
    @property
    def mapping_model(self) -> TBoxBaseMapping:
        "TBox Mapping Model"
        return TBoxBaseMapping

    # OVERRIDE
    @property
    def json_ld(self) -> "Dict[str, Any]":
        """Make the json-ld if pipeline is in abox-mode"""
        return _make_tbox_json_ld(self)

    # OVERRIDE
    @classmethod
    def _run_parser(
        cls,
        self: "ExcelTBoxParser",
        datafile: BytesIO,
        mapping: "List[TBoxBaseMapping]",
    ) -> None:
        """
        Run excel parser in tbox mode.

        Parameters
        ----------
        self : ExcelTBoxParser
            The instance of the parser.
        datafile : BytesIO
            The excel file to be parsed.
        mapping : List[TBoxBaseMapping]
            The list of mappings to be applied.

        Returns
        -------
        None
            This function does not return any value.
        """
        df = pd.read_excel(datafile, sheet_name=self.sheet)
        _make_tbox_classes(self, df, mapping)

    # OVERRIDE
    @classmethod
    def _load_data_file(cls, self: "ExcelTBoxParser") -> BytesIO:
        """Load excel file"""
        return _load_data_file(self)


class ExcelABoxParser(ABoxBaseParser):
    """
    Parses a data file of type excel in a box mode
    """

    unit_from_macro: bool = Field(
        False,
        description="When enabled, units coming from excel macros will be taken into account.",
    )

    unit_macro_location: int = Field(
        -1,
        description="Index where the marco for the unit in an excel cell might be located.",
    )

    # OVERRIDE
    mapping: Union[str, List[ABoxExcelMapping]] = Field(
        ...,
        description="""File path to the mapping file to be parsed or
        a list with the mapping.""",
    )

    # OVERRIDE
    @property
    def mapping_model(self) -> ABoxExcelMapping:
        "Mapping Model"
        return ABoxExcelMapping

    # OVERRIDE
    @property
    def json_ld(self) -> Dict[str, Any]:
        """
        Returns the JSON-LD representation of the data in ABox mode.

        The JSON-LD is constructed based on the metadata and dataframe data.
        If the file description is not suppressed, it includes the metadata and dataframe data tables.
        Otherwise, it returns a list of JSON-LD representations of the individual models.

        :return: A dictionary representing the JSON-LD data.
        """

        if not self.config.suppress_file_description:
            tables = []

            if self.general_metadata:
                meta_table = {
                    "@type": "csvw:Table",
                    "rdfs:label": "Metadata",
                    "csvw:row": [],
                }

                for mapping in self.general_metadata:
                    if isinstance(mapping, QuantityGraph):
                        row = {
                            "@type": "csvw:Row",
                            "csvw:titles": {
                                "@type": "xsd:string",
                                "@value": mapping.key,
                            },
                            "qudt:quantity": mapping.json_ld,
                        }
                        meta_table["csvw:row"].append(row)
                    elif isinstance(mapping, PropertyGraph):
                        row = {
                            "@type": "csvw:Row",
                            "csvw:titles": {
                                "@type": "xsd:string",
                                "@value": mapping.key,
                            },
                            "csvw:describes": mapping.json_ld,
                        }
                        meta_table["csvw:row"].append(row)
                    else:
                        raise TypeError(
                            f"Mapping must be of type {QuantityGraph} or {PropertyGraph}, not {type(mapping)}"
                        )
                tables += [meta_table]

            if self.dataframe_metadata:
                column_schema = {"@type": "csvw:Schema", "csvw:column": []}
                tables += [
                    {
                        "@type": "csvw:Table",
                        "rdfs:label": "Dataframe",
                        "csvw:tableSchema": column_schema,
                    }
                ]
                for idx, mapping in enumerate(self.dataframe_metadata):
                    if isinstance(mapping, QuantityGraph):
                        entity = {"qudt:quantity": mapping.json_ld}
                    elif isinstance(mapping, PropertyGraph):
                        entity = {"dcterms:subject": mapping.json_ld}
                    else:
                        raise TypeError(
                            f"Mapping must be of type {QuantityGraph} or {PropertyGraph}, not {type(mapping)}"
                        )

                    if self.config.data_download_uri:
                        download_url = {
                            "dcterms:identifier": {
                                "@type": "xsd:anyURI",
                                "@value": urljoin(
                                    str(self.config.data_download_uri),
                                    f"column-{idx}",
                                ),
                            }
                        }
                    else:
                        download_url = {}

                    column = {
                        "@type": "csvw:Column",
                        "csvw:titles": {
                            "@type": "xsd:string",
                            "@value": mapping.key,
                        },
                        **entity,
                        "foaf:page": {
                            "@type": "foaf:Document",
                            "dcterms:format": {
                                "@type": "xsd:anyURI",
                                "@value": "https://www.iana.org/assignments/media-types/application/json",
                            },
                            "dcterms:type": {
                                "@type": "xsd:anyURI",
                                "@value": "http://purl.org/dc/terms/Dataset",
                            },
                            **download_url,
                        },
                    }
                    column_schema["csvw:column"].append(column)

            # flatten list if only one value exists
            if len(tables) == 1:
                tables = tables.pop()
            # make relation to csvw:table property
            if tables:
                csvw_tables = {"csvw:table": tables}
            else:
                csvw_tables = {}

            json_ld = {
                "@context": {
                    f"{self.config.prefix_name}": make_prefix(self.config),
                    "csvw": "http://www.w3.org/ns/csvw#",
                    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                    "dcat": "http://www.w3.org/ns/dcat#",
                    "xsd": "http://www.w3.org/2001/XMLSchema#",
                    "dcterms": "http://purl.org/dc/terms/",
                    "qudt": "http://qudt.org/schema/qudt/",
                    "csvw": "http://www.w3.org/ns/csvw#",
                    "foaf": "http://xmlns.com/foaf/spec/",
                },
                "@id": f"{self.config.prefix_name}:tableGroup",
                "@type": "csvw:TableGroup",
                **csvw_tables,
            }
        else:
            json_ld = {
                "@graph": [model.json_ld for model in self.general_metadata]
                + [model.json_ld for model in self.dataframe_metadata]
            }
        return json_ld

    # OVERRIDE
    @classmethod
    def _run_parser(
        cls,
        self: "ExcelABoxParser",
        datafile: BytesIO,
        mapping: "List[ABoxExcelMapping]",
    ) -> None:
        """
        Parses the metadata, dataframe metadata, and dataframe from an Excel file.

        Args:
            self (ExcelABoxParser): The instance of the ExcelABoxParser class.
            datafile (BytesIO): The file object containing the Excel file.
            mapping (List[ABoxExcelMapping]): The list of mappings to use for parsing.

        Returns:
            None: This function does not return anything.
        """

        workbook = load_workbook(filename=datafile, data_only=True)
        datafile.seek(0)
        macros = load_workbook(filename=datafile)
        datafile.seek(0)

        self._general_metadata = []
        self._dataframe_metadata = []
        self._dataframe = {}

        for datum in mapping:
            worksheet = workbook[datum.worksheet]

            if datum.suffix_from_location:
                suffix = worksheet[datum.suffix].value
                if not suffix:
                    suffix = datum.suffix
                    message = f"""Could not properly resolve suffix location `{datum.suffix}`
                                  Will use the location itself as suffix.
                                """
                    warnings.warn(message, MappingMissmatchWarning)
            else:
                suffix = datum.suffix
            suffix = quote(suffix)

            if not datum.custom_relations:
                if datum.value_location and datum.dataframe_start:
                    raise RuntimeError(
                        """Both, `value_location` and `dataframe_start
                        are set. Only one of them must be set."""
                    )

                # find data for dataframe
                if datum.dataframe_start:
                    column_name = datum.dataframe_start.rstrip("0123456789")
                    dataframe_end = f"{column_name}{worksheet.max_row}"

                    column = worksheet[datum.dataframe_start : dataframe_end]
                    if column:
                        self._dataframe[suffix] = [
                            cell[0].value for cell in column
                        ]
                    else:
                        message = f"""Concept with key `{datum.key}`
                                    does not have a dataframe from `{datum.dataframe_start}`
                                    until `{dataframe_end}` .
                                    Concept will be omitted in graph.
                                    """
                        warnings.warn(message, MappingMissmatchWarning)

                # check if there is a macro for the unit of the entity
                if self.unit_from_macro and datum.value_location:
                    macro_worksheet = macros[datum.worksheet]
                    macro_value_cell = macro_worksheet[
                        datum.value_location
                    ].number_format.split()
                    if len(macro_value_cell) != 1:
                        macro_unit = macro_value_cell[self.unit_macro_location]
                    else:
                        macro_unit = None
                else:
                    macro_unit = None

                # check if there is a unit somewhere in the sheet
                if datum.unit_location:
                    unit_location = worksheet[datum.unit_location].value
                    if not unit_location:
                        message = f"""Concept with key `{datum.key}`
                                    does not have a unit at location `{datum.unit_location}`.
                                    This mapping for the unit will be omitted in graph.
                                    """
                        warnings.warn(message, MappingMissmatchWarning)
                else:
                    unit_location = None

                # decide which unit to take
                unit = datum.unit or unit_location or macro_unit
                if unit:
                    if not isinstance(unit, str):
                        raise TypeError(
                            f"""Unit `{unit}` for key `{datum.key}` is not a string.
                            Is it a bad mapping?"""
                        )
                    unit = _strip_unit(unit, self.config.remove_from_unit)

                # make model
                model_data = {
                    "key": datum.key,
                    "unit": unit,
                    "iri": datum.iri,
                    "suffix": suffix,
                    "annotation": datum.annotation,
                    "config": self.config,
                }

                if datum.value_location and not datum.dataframe_start:
                    value = worksheet[datum.value_location].value

                    if model_data.get("unit") and _value_exists(value):
                        model_data["value"] = value
                    elif not model_data.get("unit") and _value_exists(value):
                        model_data["value"] = str(value)
                    else:
                        message = f"""Concept with key `{datum.key}`
                                    does not have a value at location `{datum.value_location}`.
                                    Concept will be omitted in graph.
                                    """
                        warnings.warn(message, MappingMissmatchWarning)
                else:
                    value = None

                value_exists = _value_exists(value)

                if value_exists or suffix in self.dataframe:
                    if datum.value_relation:
                        model_data["value_relation"] = datum.value_relation
                    if model_data.get("unit"):
                        if datum.unit_relation:
                            model_data["unit_relation"] = datum.unit_relation
                        model = QuantityGraph(**model_data)
                    else:
                        model = PropertyGraph(
                            **model_data,
                            value_datatype=datum.value_datatype,
                            value_relation_type=datum.value_relation_type,
                        )

                    if value_exists:
                        self._general_metadata.append(model)
                    else:
                        self._dataframe_metadata.append(model)

            else:
                for relation in datum.custom_relations:
                    value = worksheet[relation.object_location].value

                    if isinstance(
                        relation.object_data_type,
                        (
                            CustomRelationPropertySubgraph,
                            CustomRelationQuantitySubgraph,
                        ),
                    ):
                        if isinstance(
                            relation.object_data_type,
                            CustomRelationPropertySubgraph,
                        ):
                            Model = PropertyGraph
                        else:
                            Model = QuantityGraph
                        model = Model(
                            value=value,
                            **relation.object_data_type.model_dump(),
                        )
                        model.suffix += "_" + suffix
                        model = PropertyGraph(
                            value_relation=relation.relation,
                            value_relation_type="object_property",
                            value=model,
                            iri=datum.iri,
                            suffix=suffix,
                            config=self.config,
                        )
                        self._general_metadata.append(model)
                    elif _value_exists(value):
                        model = PropertyGraph(
                            value_relation=relation.relation,
                            value_relation_type=relation.relation_type,
                            value_datatype=relation.object_data_type,
                            value=value,
                            iri=datum.iri,
                            suffix=suffix,
                            config=self.config,
                        )
                        self._general_metadata.append(model)
                    else:
                        message = f"""Concept with for iri `{datum.iri}`
                                        does not have a value at location `{relation.object_location}`.
                                        Concept will be omitted in graph.
                                        """
                        warnings.warn(message, MappingMissmatchWarning)

        # set dataframe as pd dataframe
        self._dataframe = pd.DataFrame.from_dict(
            self._dataframe, orient="index"
        ).transpose()
        # check if drop na:
        if self.dropna:
            self._dataframe.dropna(how="all", inplace=True)

    # OVERRIDE
    @classmethod
    def _load_data_file(cls, self: "ExcelABoxParser") -> BytesIO:
        """Load excel file"""
        return _load_data_file(self)


class ExcelParser(BaseFileParser):
    """Parser for excel files"""

    # OVERRIDE
    @property
    def _abox_parser(self) -> ExcelABoxParser:
        """Pydantic Model for Excel ABox parser"""
        return ExcelABoxParser

    # OVERRIDE
    @property
    def _tbox_parser(self) -> ExcelTBoxParser:
        """Pydantic Model for Excel TBox parser"""
        return ExcelTBoxParser

    # OVERRIDE
    @property
    def media_type(self) -> str:
        """IANA Media type definition of the resource to be parsed."""
        return "https://www.iana.org/assignments/media-types/application/vnd.ms-excel"



"""Mapping models for data2rdf"""


from typing import List, Optional, Union

from pydantic import (
    AliasChoices,
    AnyUrl,
    BaseModel,
    Field,
    field_validator,
    model_validator,
)

from .base import BasicConceptMapping, BasicSuffixModel, RelationType


class TBoxBaseMapping(BasicConceptMapping):
    """Mapping between a object/data/annotation property and
    a value under a location in the data file. This"""

    # OVERRIDE
    key: str = Field(
        ...,
        description="""Key/Column/Location of the value in the data file.
        The value can be a float/int/str or URI""",
    )
    relation: Union[str, AnyUrl] = Field(
        ...,
        description="""Object/Data/Annotation property for the value
        resolving from `key` of this model""",
    )
    relation_type: RelationType = Field(
        ..., description="Type of the semantic relation used in the mappings"
    )

    datatype: Optional[str] = Field(
        None,
        description="XSD Datatype of the targed value",
        alias=AliasChoices("datatype", "data_type"),
    )


class PropertySubgraphBaseModel(BasicSuffixModel):
    concatenate: Optional[bool] = Field(
        False,
        description="Concatenate the value and the iri",
        alias=AliasChoices("concatenate", "concat"),
    )


class CustomRelationPropertySubgraph(PropertySubgraphBaseModel):
    value_relation: Optional[str] = Field(
        "rdfs:label",
        description="""Object/Data/Annotation property for the value
        resolving from `key` of this model""",
    )


class CustomRelationQuantitySubgraph(PropertySubgraphBaseModel):
    unit_relation: Optional[Union[str, AnyUrl]] = Field(
        "qudt:hasUnit",
        description="""Object property for mapping the IRI
         of the unit to the individual.""",
    )
    value_relation: Optional[Union[str, AnyUrl]] = Field(
        "qudt:value",
        description="""Data property
        for mapping the data value to the individual.""",
    )
    unit: Optional[Union[str, AnyUrl]] = Field(
        None, description="Symbol or QUDT IRI for the mapping"
    )


class CustomRelation(BaseModel):
    """Custom relation model"""

    relation: Union[str, AnyUrl] = Field(
        ...,
        description="""Object/Data/Annotation property for the value
        resolving from `key` of this model""",
    )
    object_location: Optional[str] = Field(
        ...,
        description="Cell number or Jsonpath to the value of the quantity or property",
    )
    object_data_type: Optional[
        Union[
            str, CustomRelationPropertySubgraph, CustomRelationQuantitySubgraph
        ]
    ] = Field(
        None,
        description="XSD Data type of the object or PropertyGraph-mapping or QuantityGraph-mapping",
        alias=AliasChoices(
            "object_datatype", "object_data_type", "object_type"
        ),
    )
    relation_type: Optional[RelationType] = Field(
        None, description="Type of the semantic relation used in the mappings"
    )


class ABoxBaseMapping(BasicConceptMapping, BasicSuffixModel):
    """Base class for mapping during A Box modelling"""

    unit: Optional[Union[str, AnyUrl]] = Field(
        None, description="Symbol or QUDT IRI for the mapping"
    )
    annotation: Optional[Union[str, AnyUrl]] = Field(
        None, description="Base IRI with which the value shall be concatenated"
    )
    custom_relations: Optional[List[CustomRelation]] = Field(
        None,
        description="""In case if `value_location`, `unit_location` ,
        `value_relation` and `unit_relation` is not used, a user can also
        specify the properties of the individual produced in this custom relation
        fields.""",
    )

    source: Optional[str] = Field(
        None,
        description="""In case if the json parser is used and the `custom_relations` are set:
                                  Source and iterate over mupltiple objects from a given jsonpath, e.g. "$.data[*]".
                                  The mapping will be applied to all the iterated objects.""",
    )

    value_location: Optional[str] = Field(
        None,
        description="Cell number or Jsonpath to the value of the quantity or property",
    )
    unit_location: Optional[str] = Field(
        None, description="cell number or Jsonpath to the unit of the property"
    )
    value_relation: Optional[Union[str, AnyUrl]] = Field(
        None,
        description="""Data or annotation property
        for mapping the data value to the individual.""",
    )
    value_relation_type: Optional[RelationType] = Field(
        None, description="Type of the semantic relation used in the mappings"
    )
    value_datatype: Optional[str] = Field(
        None, description="XSD Datatype of the targed value"
    )
    unit_relation: Optional[Union[str, AnyUrl]] = Field(
        None,
        description="""Object property for mapping the IRI
         of the unit to the individual, in case the concept
         is a quantity and has a unit""",
    )
    suffix_from_location: bool = Field(
        False,
        description="When enabled, the suffix will be taken from the location, e.g. a cell number",
    )

    @field_validator("annotation", mode="after")
    @classmethod
    def validate_annotation(
        cls, value: Optional[Union[str, AnyUrl]]
    ) -> Optional[AnyUrl]:
        if not (isinstance(value, str) and len(value) == 0) or isinstance(
            value, AnyUrl
        ):
            return value

    @model_validator(mode="after")
    @classmethod
    def validate_model(cls, self: "ABoxBaseMapping") -> "ABoxBaseMapping":
        """Validate model"""
        if (
            self.value_location
            or self.unit_location
            or self.value_relation
            or self.unit_relation
        ) and self.custom_relations:
            raise ValueError(
                "value_location, unit_location, value_relation and unit_relation are mutually exclusive with custom_relations"
            )
        return self


class ABoxExcelMapping(ABoxBaseMapping):
    """A special model for mapping from excel files to semantic concepts in the ABox"""

    dataframe_start: Optional[str] = Field(
        None,
        description="Cell location for the start of the dataframe quantity",
        alias=AliasChoices("dataframe_start", "time_series_start"),
    )
    worksheet: Optional[str] = Field(
        None,
        description="Name of the worksheet where the entity is located in the excel file",
    )



"""data2rdf pytest utilty"""
import random
import string
import time


def remove_ids(metadata: dict) -> dict:
    """
    Remove 'id' keys from metadata sections and entries.

    Modifies the input metadata dictionary directly.

    :param metadata: The metadata dictionary to modify.
    :return: The input metadata dictionary, with 'id' keys removed.
    """
    for section in metadata.get("sections", []):
        section.pop("id")
        for entry in section.get("entries", []):
            entry.pop("id")
    sort_entries(metadata)
    return metadata


def sort_entries(metadata: dict) -> dict:
    """
    Sort entries in metadata by label.

    Modifies the input metadata dictionary directly.
    """
    for section in metadata.get("sections", []):
        section["entries"].sort(key=lambda x: x["label"])
    return metadata


def as_non_dsms_schema(metadata: dict) -> dict:
    """
    Convert DSMS schema to a flat dictionary.

    The input should be metadata in the DSMS schema, i.e. a dictionary
    with a "sections" key. Each section should have an "entries" key with
    a list of dictionaries. Each of these dictionaries should have a
    "label" key.

    The output is a dictionary with the same labels as keys and the same
    dictionaries as values.
    """
    response = {}
    for section in metadata.get("sections", []):
        section["entries"].sort(key=lambda x: x["label"])
        for entry in section.get("entries", []):
            response[entry["label"]] = entry
    return response


def dsms_schema(metadata: dict) -> dict:
    """
    Convert a flat dictionary to a DSMS schema.

    The input should be a dictionary with each key-value pair representing
    a metadata entry. The output is a dictionary in the DSMS schema, with
    a single section named "General", containing the given metadata entries.

    :param metadata: The metadata dictionary to convert.
    :return: A dictionary in the DSMS schema.
    """
    if metadata:
        for metadatum in metadata:
            metadatum["id"] = generate_id()
        metadata = {
            "sections": [
                {
                    "id": generate_id(),
                    "name": "General",
                    "entries": metadata,
                }
            ]
        }
    else:
        metadata = {}

    return metadata


def generate_id(prefix: str = "id") -> str:
    # Generate a unique part using time and random characters
    """
    Generates a unique id using a combination of the current time and 6 random characters.

    Args:
    prefix (str): The prefix to use for the generated id. Defaults to "id".

    Returns:
    str: The generated id.
    """
    unique_part = f"{int(time.time() * 1000)}"  # Milliseconds since epoch
    random_part = "".join(
        random.choices(string.ascii_lowercase + string.digits, k=6)  # nosec
    )
    # Combine prefix, unique part, and random part
    generated_id = f"{prefix}{unique_part}{random_part}"
    return generated_id



"""Test DSMS utilities"""

import pytest

mappings = [
    ("http://qudt.org/vocab/unit/PER-MilliM", "mm-1"),
    ("http://qudt.org/vocab/unit/MegaPA", "MPa"),
    ("http://qudt.org/vocab/unit/MilliM-PER-BAR", "mm/bar"),
    ("http://qudt.org/vocab/unit/MilliM-PER-BAR", "mm.bar-1"),
    ("http://qudt.org/vocab/unit/MilliM2", "mm²"),
    ("http://qudt.org/vocab/unit/DEG_C", "°C"),
]


@pytest.mark.parametrize("iri,symbol", mappings)
def test_unit_retrieval(iri, symbol) -> None:
    from data2rdf import Config
    from data2rdf.qudt.utils import _get_query_match

    config = Config()

    assert [iri] == _get_query_match(symbol, config.qudt_units)



"""Test custom relations for JSON parser"""
import os

EXPECTED = """
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix ns1: <https://w3id.org/emmo/domain/characterisation-methodology/chameo#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix ns2: <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#> .
@prefix ns3: <https://w3id.org/steel/ProcessOntology/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

ns2:Jane a ns1:Operator ;
            foaf:name "Jane"^^xsd:string ;
            foaf:age 28 ;
            ns3:hasLaboratory 123 .

ns2:John a ns1:Operator ;
            foaf:name "John"^^xsd:string ;
            foaf:age 32 ;
            ns3:hasLaboratory 456 .
"""

EXPECTED_DATATYPE = """
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix ns1: <https://w3id.org/emmo/domain/characterisation-methodology/chameo#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix ns2: <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#> .
@prefix ns3: <https://w3id.org/steel/ProcessOntology/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

ns2:Jane a ns1:Operator ;
            foaf:name "Jane"^^xsd:string ;
            foaf:age "28"^^xsd:string ;
            ns3:hasLaboratory "123"^^xsd:string .

ns2:John a ns1:Operator ;
            foaf:name "John"^^xsd:string ;
            foaf:age "32"^^xsd:string ;
            ns3:hasLaboratory "456"^^xsd:string .
"""


MAPPING_SUBGRAPHS = [
    {
        "iri": "https://w3id.org/steel/ProcessOntology/Specimen",
        "suffix": "A2",
        "worksheet": "tab1",
        "suffix_from_location": True,
        "custom_relations": [
            {
                "object_location": "B2",
                "relation": "https://w3id.org/steel/ProcessOntology/hasYoungsModulus",
                "object_type": {
                    "iri": "https://w3id.org/steel/ProcessOntology/YoungsModulus",
                    "unit": "GPa",
                },
            },
            {
                "object_location": "C2",
                "relation": "https://w3id.org/steel/ProcessOntology/hasMaterial",
                "object_type": {
                    "iri": "https://w3id.org/steel/ProcessOntology/Material",
                    "value_relation": "https://w3id.org/steel/ProcessOntology/hasIdentifier",
                },
            },
            {
                "object_location": "A2",
                "relation": "https://w3id.org/steel/ProcessOntology/hasIdentifier",
            },
        ],
    },
    {
        "iri": "https://w3id.org/steel/ProcessOntology/Specimen",
        "suffix": "A3",
        "worksheet": "tab1",
        "suffix_from_location": True,
        "custom_relations": [
            {
                "object_location": "B3",
                "relation": "https://w3id.org/steel/ProcessOntology/hasYoungsModulus",
                "object_type": {
                    "iri": "https://w3id.org/steel/ProcessOntology/YoungsModulus",
                    "unit": "GPa",
                },
            },
            {
                "object_location": "C3",
                "relation": "https://w3id.org/steel/ProcessOntology/hasMaterial",
                "object_type": {
                    "iri": "https://w3id.org/steel/ProcessOntology/Material",
                    "value_relation": "https://w3id.org/steel/ProcessOntology/hasIdentifier",
                },
            },
            {
                "object_location": "A3",
                "relation": "https://w3id.org/steel/ProcessOntology/hasIdentifier",
            },
        ],
    },
]

EXPECTED_SUBGRAPHS = """
@prefix ns1: <https://w3id.org/steel/ProcessOntology/> .
@prefix ns2: <http://qudt.org/schema/qudt/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#specimen_1> a ns1:Specimen ;
    ns1:hasIdentifier "specimen_1"^^xsd:string ;
    ns1:hasMaterial <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#Material_specimen_1> ;
    ns1:hasYoungsModulus <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#YoungsModulus_specimen_1> .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#specimen_2> a ns1:Specimen ;
    ns1:hasIdentifier "specimen_2"^^xsd:string ;
    ns1:hasMaterial <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#Material_specimen_2> ;
    ns1:hasYoungsModulus <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#YoungsModulus_specimen_2> .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#Material_specimen_1> a ns1:Material ;
    ns1:hasIdentifier "material_1"^^xsd:string .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#Material_specimen_2> a ns1:Material ;
    ns1:hasIdentifier "material_2"^^xsd:string .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#YoungsModulus_specimen_1> a ns1:YoungsModulus ;
    ns2:hasUnit "http://qudt.org/vocab/unit/GigaPA"^^xsd:anyURI ;
    ns2:value 100 .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#YoungsModulus_specimen_2> a ns1:YoungsModulus ;
    ns2:hasUnit "http://qudt.org/vocab/unit/GigaPA"^^xsd:anyURI ;
    ns2:value 200 .

"""


MAPPING_INDEX = [
    {
        "iri": "https://w3id.org/emmo/domain/characterisation-methodology/chameo#Operator",
        "suffix": "A2",
        "suffix_from_location": True,
        "worksheet": "tab1",
        "custom_relations": [
            {
                "object_location": "A2",
                "relation": "http://xmlns.com/foaf/0.1/name",
            },
            {
                "object_location": "B2",
                "relation": "http://xmlns.com/foaf/0.1/age",
            },
            {
                "object_location": "C2",
                "relation": "https://w3id.org/steel/ProcessOntology/hasLaboratory",
            },
        ],
    },
    {
        "iri": "https://w3id.org/emmo/domain/characterisation-methodology/chameo#Operator",
        "suffix": "A3",
        "suffix_from_location": True,
        "worksheet": "tab1",
        "custom_relations": [
            {
                "object_location": "A3",
                "relation": "http://xmlns.com/foaf/0.1/name",
            },
            {
                "object_location": "B3",
                "relation": "http://xmlns.com/foaf/0.1/age",
            },
            {
                "object_location": "C3",
                "relation": "https://w3id.org/steel/ProcessOntology/hasLaboratory",
            },
        ],
    },
]


MAPPING_INDEX_DATATYPE = [
    {
        "iri": "https://w3id.org/emmo/domain/characterisation-methodology/chameo#Operator",
        "suffix": "A2",
        "suffix_from_location": True,
        "worksheet": "tab1",
        "custom_relations": [
            {
                "object_location": "A2",
                "relation": "http://xmlns.com/foaf/0.1/name",
                "object_data_type": "string",
            },
            {
                "object_location": "B2",
                "relation": "http://xmlns.com/foaf/0.1/age",
                "object_data_type": "string",
            },
            {
                "object_location": "C2",
                "relation": "https://w3id.org/steel/ProcessOntology/hasLaboratory",
                "object_data_type": "string",
            },
        ],
    },
    {
        "iri": "https://w3id.org/emmo/domain/characterisation-methodology/chameo#Operator",
        "suffix": "A3",
        "suffix_from_location": True,
        "worksheet": "tab1",
        "custom_relations": [
            {
                "object_location": "A3",
                "relation": "http://xmlns.com/foaf/0.1/name",
                "object_data_type": "string",
            },
            {
                "object_location": "B3",
                "relation": "http://xmlns.com/foaf/0.1/age",
                "object_data_type": "string",
            },
            {
                "object_location": "C3",
                "relation": "https://w3id.org/steel/ProcessOntology/hasLaboratory",
                "object_data_type": "string",
            },
        ],
    },
]

BASE_IRI = (
    "https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation"
)

test_folder = os.path.dirname(os.path.abspath(__file__))
DATA = os.path.join(test_folder, "data.xlsx")
DATA2 = os.path.join(test_folder, "data2.xlsx")


def test_pipeline_xlsx_custom_relations() -> None:
    from rdflib import Graph

    from data2rdf import Data2RDF, Parser

    pipeline = Data2RDF(
        raw_data=DATA,
        mapping=MAPPING_INDEX,
        parser=Parser.excel,
        parser_args={"dropna": True, "unit_from_macro": True},
        config={
            "base_iri": BASE_IRI,
            "separator": "#",
            "prefix_name": "nanoindentation",
            "suppress_file_description": True,
        },
    )
    expected_graph = Graph()
    expected_graph.parse(data=EXPECTED)

    assert pipeline.graph.isomorphic(expected_graph)


def test_pipeline_xlsx_custom_relations_datatype() -> None:
    """Test with custom relations and datatypes"""

    from rdflib import Graph

    from data2rdf import Data2RDF, Parser

    pipeline = Data2RDF(
        raw_data=DATA,
        mapping=MAPPING_INDEX_DATATYPE,
        parser=Parser.excel,
        parser_args={"dropna": True, "unit_from_macro": True},
        config={
            "base_iri": BASE_IRI,
            "separator": "#",
            "prefix_name": "nanoindentation",
            "suppress_file_description": True,
        },
    )

    expected_graph = Graph()
    expected_graph.parse(data=EXPECTED_DATATYPE)

    assert pipeline.graph.isomorphic(expected_graph)


def test_pipeline_xlsx_custom_relations_subgraphs() -> None:
    """Test with custom relations and datatypes"""

    from rdflib import Graph

    from data2rdf import Data2RDF, Parser

    pipeline = Data2RDF(
        raw_data=DATA2,
        mapping=MAPPING_SUBGRAPHS,
        parser=Parser.excel,
        parser_args={"dropna": True, "unit_from_macro": False},
        config={
            "base_iri": BASE_IRI,
            "separator": "#",
            "suppress_file_description": True,
        },
    )

    expected_graph = Graph()
    expected_graph.parse(data=EXPECTED_SUBGRAPHS)

    assert pipeline.graph.isomorphic(expected_graph)



"""Data2RDF parser utilities"""

import json
import warnings
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from data2rdf import Config
from data2rdf.models.graph import ClassTypeGraph
from data2rdf.warnings import MappingMissmatchWarning

from data2rdf.models.mapping import (  # isort:skip
    RelationType,
)


if TYPE_CHECKING:
    from typing import Any, Dict, List, Union

    from data2rdf.models.mapping import TBoxBaseMapping
    from data2rdf.parsers.base import TBoxBaseParser


def load_mapping_file(
    mapping: "Union[str, List[Dict[str, Any]]]",
    config: "Config" = Config(),
    model_callable: "Any" = dict,
) -> "List[Any]":
    """
    Load a mapping file and transform its contents into a dictionary format.

    Parameters:
        mapping (Union[str, List[Dict[str, Any]]]): Path to the mapping file (either a string representing the file path or a list containing the mapping directly).
        config (Config, optional): Configuration settings for loading the file. Defaults to Config().
        model_callable (Any, optional): Callable object used to transform each row of the mapping file into the desired format. Defaults to dict.

    Returns:
        List[Any]: A list containing the loaded mapping data.

    Raises:
        TypeError: If the `mapping` parameter is not of type `str` or `dict`, or if the file type for mapping is not supported.

    Note:
        - For Excel files (.xlsx), the 'sameas' sheet is read from the Excel file.
        The contents are then transformed into a dictionary where each row corresponds to a key-value pair.
        - For CSV files (.csv) please take care of the correct separator to be set in the config under `mapping_csv_separator`.
        - For JSON files (.json), the entire file is loaded as a dictionary.
        - If `mapping` is already a dictionary, it is returned as is.

    """
    if not isinstance(mapping, (str, list)):
        raise TypeError(
            f"""Mapping file must be of type `{str}` or `{list}`,
            not `{type(mapping)}`."""
        )
    if isinstance(mapping, str):
        if mapping.endswith("xlsx"):
            mapping_df = pd.read_excel(
                mapping,
                sheet_name="sameas",
                engine="openpyxl",
            )
            mapping_df.fillna("", inplace=True)
            mapping_df = mapping_df.apply(lambda s: s.str.replace('"', ""))
            model = [row.to_dict() for n, row in mapping_df.iterrows()]

        elif mapping.endswith("json"):
            with open(mapping, encoding=config.encoding) as file:
                model = json.load(file)
        elif mapping.endswith("csv"):
            mapping_df = pd.read_csv(
                mapping,
                sep=config.mapping_csv_separator,
                encoding=config.encoding,
            )
            mapping_df.fillna("", inplace=True)
            mapping_df = mapping_df.apply(lambda s: s.str.replace('"', ""))
            model = [row.to_dict() for n, row in mapping_df.iterrows()]
        else:
            raise TypeError("File type for mapping not supported!")

        result = [model_callable(**row) for row in model]
    if isinstance(mapping, list):
        result = mapping
    return result


def _strip_unit(symbol: str, char_list: "List[str]") -> str:
    for char in char_list:
        symbol = symbol.strip(char)
    return symbol


def _make_tbox_classes(
    self: "TBoxBaseParser",
    df: "pd.DataFrame",
    mapping: "List[TBoxBaseMapping]",
) -> None:
    self._classes = []
    mapping = {model.key: model for model in mapping}
    if hasattr(self, "header_length"):
        skipped = df[self.header_length - 1 :]
    else:
        skipped = df
    for n, row in skipped.iterrows():
        annotations = []
        datatypes = []
        objects = []

        for key, model in mapping.items():
            try:
                value = row[key]
                if isinstance(value, float) and np.isnan(value):
                    value = self.fillna
                if value:
                    relation_mapping = {
                        "value": value,
                        "relation": model.relation,
                        "datatype": model.datatype,
                    }
                    if model.relation_type == RelationType.ANNOTATION_PROPERTY:
                        annotations.append(relation_mapping)
                    if model.relation_type == RelationType.DATA_PROPERTY:
                        datatypes.append(relation_mapping)
                    if model.relation_type == RelationType.OBJECT_PROPERTY:
                        objects.append(relation_mapping)
                else:
                    warnings.warn(
                        f"Data for key `{key}` does not exist in row {n}.",
                        MappingMissmatchWarning,
                    )
            except KeyError:
                warnings.warn(
                    f"Data for key `{key}` does not exist in row {n}.",
                    MappingMissmatchWarning,
                )

        if self.rdfs_type_location:
            rdfs_type = row[self.rdfs_type_location]
            if isinstance(rdfs_type, type(None)) or (
                isinstance(rdfs_type, float) and np.isnan(rdfs_type)
            ):
                warnings.warn(
                    f"Data for key `{self.rdfs_type_location}` does not exist in row {n}.",
                    MappingMissmatchWarning,
                )
                rdfs_type = "owl:Class"
        else:
            rdfs_type = "owl:Class"
        subgraph = ClassTypeGraph(
            rdfs_type=rdfs_type,
            suffix=row[self.suffix_location],
            annotation_properties=annotations,
            object_properties=objects,
            data_properties=datatypes,
            config=self.config,
        )
        self._classes.append(subgraph)


def _make_tbox_json_ld(model: "TBoxBaseParser") -> "Dict[str, Any]":
    ontology_iri = model.ontology_iri or model.config.base_iri
    classes = [model.json_ld for model in model.classes]
    if not model.config.exclude_ontology_title:
        if model.ontology_title:
            title = {"dcterms:title": model.ontology_title}
        else:
            title = {}
        if model.authors:
            authors = {
                "dcterms:creator": [
                    {"@type": "foaf:Person", "foaf:name": author}
                    for author in model.authors
                ]
            }
        else:
            authors = {}
        if model.version_info:
            version = {"owl:versionInfo": model.version_info}
        else:
            version = {}

        classes += [
            {
                "@id": str(ontology_iri),
                "@type": "owl:Ontology",
                **title,
                **authors,
                **version,
            },
        ]
    return {
        "@context": {
            "owl": "http://www.w3.org/2002/07/owl#",
            "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
            "dcterms": "http://purl.org/dc/terms/",
            "foaf": "http://xmlns.com/foaf/spec/",
        },
        "@graph": classes,
    }


def _value_exists(value: "Any") -> bool:
    """
    Check if a given value exists and is valid.

    This method verifies if the input value is neither None, an empty string,
    nor a NaN (Not a Number) value.

    Args:
        value (Any): The value to be checked.

    Returns:
        bool: True if the value exists and is valid, otherwise False.
    """
    return pd.notnull(value) and value != ""



"""JSON Parser pytest"""

import json
import os

import pytest

from ..utils import dsms_schema, remove_ids, sort_entries

test_folder = os.path.dirname(os.path.abspath(__file__))
working_folder = os.path.join(test_folder, "input")
output_folder = os.path.join(test_folder, "output")

mapping_folder = os.path.join(working_folder, "mapping")
mapping_file = os.path.join(mapping_folder, "tensile_test_mapping.json")
raw_data_file = os.path.join(working_folder, "data", "sample_data.json")
expected = os.path.join(output_folder, "output_json_parser.ttl")

metadata = {
    "sections": [
        {
            "entries": [
                {
                    "label": "Remark",
                    "value": "foobar",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/Remark"
                    },
                },
                {
                    "label": "WidthChange",
                    "measurement_unit": {
                        "iri": "http://qudt.org/vocab/unit/MilliM",
                        "label": "Millimetre",
                        "namespace": "http://qudt.org/vocab/unit",
                        "symbol": "mm",
                    },
                    "value": 1.0,
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/WidthChange"
                    },
                },
            ],
            "name": "General",
        },
    ],
}

series = {"PercentageElongation": [1.0, 2.0, 3.0], "Force": [2.0, 3.0, 4.0]}


@pytest.mark.parametrize(
    "mapping_format, data_format",
    [(dict, str), (str, str), (dict, dict), (str, dict)],
)
def test_parser_json(mapping_format, data_format) -> None:
    from rdflib import Graph

    from data2rdf.models import PropertyGraph, QuantityGraph
    from data2rdf.parsers import JsonParser

    if mapping_format == str:
        mapping = mapping_file
    else:
        with open(mapping_file, encoding="utf-8") as file:
            mapping = json.load(file)

    if data_format == str:
        raw_data = raw_data_file
    else:
        with open(raw_data_file, encoding="utf-8") as file:
            raw_data = json.load(file)

    parser = JsonParser(raw_data=raw_data, mapping=mapping)

    assert len(parser.general_metadata) == 2
    for row in parser.general_metadata:
        assert isinstance(row, QuantityGraph) or isinstance(row, PropertyGraph)

    assert len(parser.dataframe_metadata) == 2
    for row in parser.dataframe_metadata:
        assert isinstance(row, QuantityGraph)

    assert len(parser.dataframe.columns) == 2
    assert sorted(series) == sorted(parser.dataframe)
    for name, column in parser.dataframe.items():
        assert len(column) == 3

    expected_graph = Graph()
    expected_graph.parse(expected)

    assert parser.graph.isomorphic(expected_graph)

    assert remove_ids(parser.to_dict(schema=dsms_schema)) == sort_entries(
        metadata
    )


@pytest.mark.parametrize("extension", ["xlsx", "json", "csv", dict])
def test_json_parser_different_mapping_files(extension) -> None:
    from rdflib import Graph

    from data2rdf.models import PropertyGraph, QuantityGraph
    from data2rdf.parsers import JsonParser

    if isinstance(extension, str):
        mapping = os.path.join(
            mapping_folder, f"tensile_test_mapping.{extension}"
        )

    else:
        path = os.path.join(mapping_folder, "tensile_test_mapping.json")
        with open(path, encoding="utf-8") as file:
            mapping = json.load(file)

    parser = JsonParser(raw_data=raw_data_file, mapping=mapping)

    assert len(parser.general_metadata) == 2
    for row in parser.general_metadata:
        assert isinstance(row, QuantityGraph) or isinstance(row, PropertyGraph)

    assert len(parser.dataframe_metadata) == 2
    for row in parser.dataframe_metadata:
        assert isinstance(row, QuantityGraph)

    assert len(parser.dataframe.columns) == 2
    assert sorted(series) == sorted(parser.dataframe)
    for name, column in parser.dataframe.items():
        assert len(column) == 3

    expected_graph = Graph()
    expected_graph.parse(expected)

    assert parser.graph.isomorphic(expected_graph)

    assert remove_ids(parser.to_dict(schema=dsms_schema)) == sort_entries(
        metadata
    )



"""Test for explicit datatypes."""

import pytest

DATA = [
    {
        "Ontological concept ID": "TestingMachine",
        "Label": "Testing machine",
        "Description": "Some description",
        "Superclass": "http://example.org#Object",
        "Comment": None,
        "Source": 123,
        "Author's name": None,
        "Author's email": None,
    },
    {
        "Ontological concept ID": "hasTestingMachine",
        "Label": "has Testing machine",
        "Type": "owl:ObjectProperty",
        "Description": "Some description",
        "Comment": None,
        "Source": None,
        "Author's name": None,
        "Author's email": None,
    },
]


MAPPING = [
    {
        "key": "Label",
        "relation": "http://www.w3.org/2000/01/rdf-schema#label",
        "relation_type": "annotation_property",
    },
    {
        "key": "Description",
        "relation": "http://purl.org/dc/terms/description",
        "relation_type": "data_property",
    },
    {
        "key": "Superclass",
        "relation": "http://www.w3.org/2000/01/rdf-schema#subClassOf",
        "relation_type": "object_property",
    },
    {
        "key": "Comment",
        "relation": "http://www.w3.org/2000/01/rdf-schema#comment",
        "relation_type": "data_property",
    },
    {
        "key": "Source",
        "relation": "http://purl.org/dc/terms/source",
        "relation_type": "data_property",
        "datatype": "float",
    },
    {
        "key": "Author's name",
        "relation": "http://purl.org/dc/terms/contributor",
        "relation_type": "data_property",
    },
    {
        "key": "Author's email",
        "relation": "http://xmlns.com/foaf/0.1/mbox",
        "relation_type": "data_property",
    },
]

EXPECTED = """
@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix foaf1: <http://xmlns.com/foaf/spec/> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

<https://w3id.org/dimat> a owl:Ontology ;
    dcterms:creator [ a foaf1:Person ;
            foaf1:name "Jane Doe" ] ;
    dcterms:title "Test Ontology" ;
    owl:versionInfo "1.0.0" .

<https://w3id.org/dimat/TestingMachine> a owl:Class ;
    rdfs:label "Testing machine"^^xsd:string ;
    dcterms:description "Some description"^^xsd:string ;
    dcterms:source "123.0"^^xsd:float ;
    rdfs:subClassOf <http://example.org#Object> .

<https://w3id.org/dimat/hasTestingMachine> a owl:ObjectProperty ;
    rdfs:label "has Testing machine"^^xsd:string ;
    dcterms:description "Some description"^^xsd:string ."""


def test_explicit_datatypes():
    from rdflib import Graph

    from data2rdf.warnings import MappingMissmatchWarning

    from data2rdf import (  # isort:skip
        Data2RDF,
        Parser,
    )

    with pytest.warns(
        MappingMissmatchWarning, match="Data for key"
    ) as warnings:
        pipeline = Data2RDF(
            mode="tbox",
            raw_data=DATA,
            mapping=MAPPING,
            parser=Parser.json,
            parser_args={
                "suffix_location": "Ontological concept ID",
                "rdfs_type_location": "Type",
                "ontology_title": "Test Ontology",
                "authors": ["Jane Doe"],
                "version_info": "1.0.0",
            },
            config={
                "base_iri": "https://w3id.org/dimat",
            },
        )

    missmatches = [
        warning
        for warning in warnings
        if warning.category == MappingMissmatchWarning
    ]
    assert len(missmatches) == 9

    expected_graph = Graph()
    expected_graph.parse(data=EXPECTED)

    assert pipeline.graph.isomorphic(expected_graph)



"""Data2RDF ABox pipeline"""

import json
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union

from rdflib import Graph

from data2rdf.config import Config
from data2rdf.modes import PipelineMode
from data2rdf.parsers import Parser
from data2rdf.utils import make_prefix

from pydantic import (  # isort:skip
    BaseModel,
    ConfigDict,
    Field,
    field_validator,
    model_validator,
)


if TYPE_CHECKING:
    from data2rdf import BasicConceptMapping


class Data2RDF(BaseModel):

    """
    Data2rdf pipeline.

    Parameters:
    - raw_data (Union[str, bytes, Dict[str, Any]]):
        In case of a csv: `str` with the file path or the content of the file itself.
        In case of a json file: `dict` for the content of the file of `str` for the file content or file path.
        In case of an excel file: `btyes` for the content or `str` for the file path
    - mapping (Union[str, Dict[str, Any]]): File path to the mapping file to be parsed or a dictionary with the mapping.
    - parser (Parser): Parser to be used depending on the type of raw data file.
    - parser_args (Dict[str, Any]): A dictionary with specific arguments for the parser. These are passed to the parser
    as keyword arguments.
    - config (Union[Dict[str, Any], Config]): Configuration object. Defaults to a new instance of Config.
    - additional_triples (Optional[Union[str, Graph]]): File path or rdflib-object for a Graph with extra triples for the
    resulting pipeline graph.
    """

    mode: PipelineMode = Field(
        PipelineMode.ABOX,
        description="""With data2rdf, you are able to model in two modes:
        the ABox (for data instances) or the TBox (for the class hierarchy/taxonomy)""",
    )

    raw_data: Union[str, bytes, Dict[str, Any], List[Dict[str, Any]]] = Field(
        ...,
        description="""
        In case of a csv: `str` with the file path or the content of the file itself.
        In case of a json file: `dict` or `list` for the content of the file of `str` for the file content or file path.
        In case of an excel file: `btyes` for the content or `str` for the file path""",
    )
    mapping: Union[str, List[Any]] = Field(
        ...,
        description="""File path to the mapping file to be parsed or
        a dictionary with the mapping.""",
    )

    parser: Parser = Field(
        ...,
        description="Parser to be used depending on the type of raw data file. ",
    )

    parser_args: Dict[str, Any] = Field(
        {},
        description="A dict with specific arguments for the parser. Is passed to the parser as kwargs.",
    )

    config: Union[Dict[str, Any], Config] = Field(
        default_factory=Config, description="Configuration object"
    )

    additional_triples: Optional[Union[str, Graph]] = Field(
        None,
        description="Filepath or rdflib-object for a Graph with extra triples for the resulting pipeline graph.",
    )

    model_config = ConfigDict(
        arbitrary_types_allowed=True, use_enum_values=True
    )

    @field_validator("config")
    @classmethod
    def validate_config(cls, value: Union[Dict[str, Any], Config]) -> Config:
        """Validate configuration"""
        if isinstance(value, dict):
            value = Config(**value)
        return value

    def _validate_additional_triples(
        self,
        value: Union[str, Graph],
    ) -> Graph:
        """Validate extra triples."""
        if isinstance(value, str):
            potential_path = Path(value)
            if potential_path.is_file():
                with open(value, encoding=self.config.encoding) as file:
                    additional_triples = file.read()
            else:
                additional_triples = value
        elif isinstance(value, Graph):
            additional_triples = value.serialize()
        else:
            raise TypeError(
                f"`additional_triples` must be of type {str}, {Graph} or {type(None)}, not {type(value)}."
            )

        additional_triples = additional_triples.replace(
            self.config.namespace_placeholder,
            make_prefix(self.config),
        )
        graph = Graph(identifier=self.config.graph_identifier)
        graph.parse(data=additional_triples)

        return graph

    @model_validator(mode="after")
    @classmethod
    def run_pipeline(cls, self: "Data2RDF") -> "Data2RDF":
        """Run pipeline."""
        self.parser = self.parser(
            raw_data=self.raw_data,
            mapping=self.mapping,
            config=self.config,
            mode=self.mode,
            parser_args=self.parser_args,
        )

        return self

    @property
    def json_ld(self) -> Dict[str, Any]:
        """
        Returns a dictionary of JSON-LD for the graph based on the pipeline mode.

        If the pipeline mode is ABOX, it returns a dictionary containing the context,
        id, type, and distribution information of the dataset. If the
        `suppress_file_description` config is False, it also includes the file
        description. Otherwise, it returns the JSON-LD of the ABox parser.

        If the pipeline mode is TBOX, it returns the JSON-LD of the TBox parser.

        Args:
            None

        Returns:
            Dict[str, Any]: A dictionary of JSON-LD for the graph.
        """

        if self.mode == PipelineMode.ABOX:
            if not self.config.suppress_file_description:
                model = {
                    "@context": {
                        f"{self.config.prefix_name}": make_prefix(self.config),
                        "csvw": "http://www.w3.org/ns/csvw#",
                        "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                        "dcat": "http://www.w3.org/ns/dcat#",
                        "xsd": "http://www.w3.org/2001/XMLSchema#",
                        "dcterms": "http://purl.org/dc/terms/",
                        "qudt": "http://qudt.org/schema/qudt/",
                        "csvw": "http://www.w3.org/ns/csvw#",
                        "foaf": "http://xmlns.com/foaf/spec/",
                    },
                    "@id": f"{self.config.prefix_name}:dataset",
                    "@type": "dcat:Dataset",
                    "dcat:distribution": {
                        "@type": "dcat:Distribution",
                        "dcat:mediaType": {
                            "@type": "xsd:anyURI",
                            "@value": self.parser.media_type,
                        },
                        "dcat:accessURL": {
                            "@type": "xsd:anyURI",
                            "@value": str(self.config.data_download_uri),
                        },
                    },
                    "dcterms:hasPart": self.parser.abox.json_ld,
                }
            else:
                model = self.parser.abox.json_ld
        elif self.mode == PipelineMode.TBOX:
            model = self.parser.tbox.json_ld
        else:
            raise TypeError("Pipeline mode not understood")
        return model

    @property
    def graph(self) -> Graph:
        """
        Returns a graph object based on the pipeline's JSON-LD data.

        The graph object is created with the identifier specified through the pipeline.
        It is then populated with the JSON-LD data from the pipeline, and if additional
        triples are provided, they are validated and added to the graph.

        Returns:
            Graph: A graph object containing the pipeline's data.
        """

        graph = Graph(identifier=self.config.graph_identifier)
        graph.parse(data=json.dumps(self.json_ld), format="json-ld")
        if self.additional_triples:
            graph += self._validate_additional_triples(self.additional_triples)
        return graph

    def to_dict(self, schema: Callable = None) -> "List[Dict[str, Any]]":
        """Return list of general metadata as DSMS custom properties"""
        if self.mode == PipelineMode.ABOX:
            return self.parser.abox.to_dict(schema=schema)
        else:
            raise NotImplementedError(
                "`to_dict()` is not available in `tbox`-mode."
            )

    @property
    def plain_metadata(self) -> Dict[str, Any]:
        """Metadata as flat json - without units and iris.
        Useful e.g. for the custom properties of the DSMS."""
        if self.mode == PipelineMode.ABOX:
            return self.parser.abox.plain_metadata
        else:
            raise NotImplementedError(
                "`plain_metadata` is not available in `tbox`-mode."
            )

    @property
    def general_metadata(self) -> "List[BasicConceptMapping]":
        """Return list object with general metadata"""
        if self.mode == PipelineMode.ABOX:
            return self.parser.abox.general_metadata
        else:
            raise NotImplementedError(
                "`general_metadata` is not available in `tbox`-mode."
            )

    @property
    def dataframe_metadata(self) -> "List[BasicConceptMapping]":
        """Return list object with dataframe metadata"""
        if self.mode == PipelineMode.ABOX:
            return self.parser.abox.dataframe_metadata
        else:
            raise NotImplementedError(
                "`dataframe_metadata` is not available in `tbox`-mode."
            )

    @property
    def dataframe(self) -> "Dict[str, Any]":
        """Return dataframe"""
        if self.mode == PipelineMode.ABOX:
            return self.parser.abox.dataframe
        else:
            raise NotImplementedError(
                "`dataframe` is not available in `tbox`-mode."
            )

    @property
    def time_series(self) -> "Dict[str, Any]":
        warnings.warn(
            "`time_series` is deprecated and will be removed in a future version. "
            "Use `dataframe` instead.",
            DeprecationWarning,
        )
        return self.dataframe

    @property
    def time_series_metadata(self) -> "List[BasicConceptMapping]":
        warnings.warn(
            "`time_series_metadata` is deprecated and will be removed in a future version. "
            "Use `dataframe_metadata` instead.",
            DeprecationWarning,
        )
        return self.dataframe_metadata



"""Test abox with multiple IRIs"""

DATA = {
    "Identifier": "123",
    "Material": [
        "https://example.materials-data.space/knowledge/material/material1",
        "https://example.materials-data.space/knowledge/material/material2",
    ],
}


MAPPING = [
    {
        "iri": [
            "https://w3id.org/emmo#EMMO_ee0466e4_780d_4236_8281_ace7ad3fc5d2",
            "https://example.org/0034d610-959d-4dd0-97f0-f3b36eeea5c4",
        ],
        "suffix": "Identifier",
        "value_location": "Identifier",
        "value_relation": "https://w3id.org/emmo#EMMO_a592c856_4103_43cf_8635_1982a1e5d5de",
        "datatype": "integer",
    },
    {
        "iri": [
            "https://w3id.org/emmo#Material",
            "https://example.org/0034d610-959d-4dd0-97f0-f3b36eeea5c4",
        ],
        "suffix": "Material",
        "value_location": "Material",
        "value_relation": "http://www.w3.org/2000/01/rdf-schema#relatedTo",
        "value_relation_type": "object_property",
    },
]

EXPECTED = """
@prefix ns1: <https://w3id.org/emmo#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

<https://example.materials-data.space/knowledge/testing-machine/Identifier> a <https://example.org/0034d610-959d-4dd0-97f0-f3b36eeea5c4>,
        ns1:EMMO_ee0466e4_780d_4236_8281_ace7ad3fc5d2 ;
    ns1:EMMO_a592c856_4103_43cf_8635_1982a1e5d5de 123 .

<https://example.materials-data.space/knowledge/testing-machine/Material> a <https://example.org/0034d610-959d-4dd0-97f0-f3b36eeea5c4>,
        ns1:Material ;
    rdfs:relatedTo <https://example.materials-data.space/knowledge/material/material1>,
        <https://example.materials-data.space/knowledge/material/material2> .

"""

DATA_CUMSTOM_RELATIONS = {
    "testing-machine": [
        {
            "Description": "Some description",
            "Identifier": "123",
            "Material": [
                "https://example.materials-data.space/knowledge/material/material1",
                "https://example.materials-data.space/knowledge/material/material2",
            ],
            "id": "0034d610-959d-4dd0-97f0-f3b36eeea5c4",
            "slug": "testing-machine1",
            "Name": "Testing machine1",
        }
    ]
}

MAPPING_CUSTOM_RELATIONS = [
    {
        "iri": [
            "https://w3id.org/emmo#EMMO_ee0466e4_780d_4236_8281_ace7ad3fc5d2",
            "https://example.org/0034d610-959d-4dd0-97f0-f3b36eeea5c4",
        ],
        "suffix": "slug",
        "source": "testing-machine[*]",
        "suffix_from_location": True,
        "custom_relations": [
            {
                "object_location": "Description",
                "relation": "http://www.w3.org/2000/01/rdf-schema#comment",
                "relation_type": "data_property",
            },
            {
                "object_location": "Identifier",
                "relation": "https://w3id.org/emmo#EMMO_a592c856_4103_43cf_8635_1982a1e5d5de",
                "relation_type": "data_property",
            },
            {
                "object_location": "Material",
                "relation": "http://www.w3.org/2000/01/rdf-schema#relatedTo",
                "relation_type": "object_property",
            },
            {
                "object_location": "Name",
                "relation": "http://www.w3.org/2000/01/rdf-schema#label",
                "relation_type": "annotation_property",
            },
        ],
    }
]

EXPECTED_CUSTOM_RELATIONS = """
@prefix ns1: <https://w3id.org/emmo#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix ex: <https://example.org/> .

<https://example.materials-data.space/knowledge/testing-machine/testing-machine1> a ns1:EMMO_ee0466e4_780d_4236_8281_ace7ad3fc5d2 , ex:0034d610-959d-4dd0-97f0-f3b36eeea5c4 ;
    rdfs:label "Testing machine1"^^xsd:string ;
    rdfs:comment "Some description"^^xsd:string ;
    rdfs:relatedTo <https://example.materials-data.space/knowledge/material/material1> ,
                   <https://example.materials-data.space/knowledge/material/material2> ;
    ns1:EMMO_a592c856_4103_43cf_8635_1982a1e5d5de 123 .
"""


def test_multiple_iris_custom_relations():
    from rdflib import Graph

    from data2rdf import Data2RDF, Parser

    pipeline = Data2RDF(
        raw_data=DATA_CUMSTOM_RELATIONS,
        mapping=MAPPING_CUSTOM_RELATIONS,
        parser=Parser.json,
        parser_args={"expand_array": True},
        config={
            "base_iri": "https://example.materials-data.space/knowledge/testing-machine/",
            "suppress_file_description": True,
        },
    )
    expected_graph = Graph()
    expected_graph.parse(data=EXPECTED_CUSTOM_RELATIONS)

    assert pipeline.graph.isomorphic(expected_graph)


def test_multiple_iris():
    from rdflib import Graph

    from data2rdf import Data2RDF, Parser

    pipeline = Data2RDF(
        raw_data=DATA,
        mapping=MAPPING,
        parser=Parser.json,
        parser_args={"expand_array": True},
        config={
            "base_iri": "https://example.materials-data.space/knowledge/testing-machine/",
            "suppress_file_description": True,
        },
    )
    expected_graph = Graph()
    expected_graph.parse(data=EXPECTED)

    assert pipeline.graph.isomorphic(expected_graph)



"""JSON Pipeline tbox pytest without title"""

import json
import os

import pytest

test_folder = os.path.dirname(os.path.abspath(__file__))
working_folder = os.path.join(test_folder, "input")
output_folder = os.path.join(test_folder, "output")

mapping_folder = os.path.join(working_folder, "mapping")
raw_data = os.path.join(working_folder, "data", "classes.json")
expected = os.path.join(output_folder, "output_json_parser_wo_title.ttl")


def test_json_pipeline_tbox_wo_title() -> None:
    from rdflib import Graph

    from data2rdf.warnings import MappingMissmatchWarning

    from data2rdf import Data2RDF, Parser  # isort:skip

    path = os.path.join(mapping_folder, "mapping.json")
    with open(path, encoding="utf-8") as file:
        mapping = json.load(file)

    with pytest.warns(
        MappingMissmatchWarning, match="Data for key"
    ) as warnings:
        pipeline = Data2RDF(
            mode="tbox",
            raw_data=raw_data,
            mapping=mapping,
            parser=Parser.json,
            parser_args={
                "suffix_location": "Ontological concept ID",
            },
            config={
                "base_iri": "https://w3id.org/dimat",
                "exclude_ontology_title": True,
            },
        )

    missmatches = [
        warning
        for warning in warnings
        if warning.category == MappingMissmatchWarning
    ]
    assert len(missmatches) == 1

    expected_graph = Graph()
    expected_graph.parse(expected)

    assert pipeline.graph.isomorphic(expected_graph)



"""CSV Parser pytest"""

import json
import os

import pytest

from ..utils import as_non_dsms_schema, dsms_schema, remove_ids, sort_entries

test_folder = os.path.dirname(os.path.abspath(__file__))
working_folder = os.path.join(test_folder, "input")
output_folder = os.path.join(test_folder, "output")

mapping_folder = os.path.join(working_folder, "mapping")
raw_data = os.path.join(working_folder, "data", "AFZ1-Fz-S1Q.xlsm")
expected = os.path.join(output_folder, "output_excel_parser.ttl")

metadata = {
    "sections": [
        {
            "entries": [
                {
                    "label": "TimeStamp",
                    "value": "2016-10-11 00:00:00",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/TimeStamp"
                    },
                },
                {
                    "label": "OriginalGaugeLength",
                    "measurement_unit": {
                        "iri": "http://qudt.org/vocab/unit/MilliM",
                        "label": "Millimetre",
                        "namespace": "http://qudt.org/vocab/unit",
                        "symbol": "mm",
                    },
                    "value": 15,
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/OriginalGaugeLength"
                    },
                },
                {
                    "label": "SpecimenWidth",
                    "measurement_unit": {
                        "iri": "http://qudt.org/vocab/unit/MilliM",
                        "label": "Millimetre",
                        "namespace": "http://qudt.org/vocab/unit",
                        "symbol": "mm",
                    },
                    "value": 9.5,
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/SpecimenWidth"
                    },
                },
                {
                    "label": "SpecimenThickness",
                    "measurement_unit": {
                        "iri": "http://qudt.org/vocab/unit/MilliM",
                        "label": "Millimetre",
                        "namespace": "http://qudt.org/vocab/unit",
                        "symbol": "mm",
                    },
                    "value": 1.5,
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/SpecimenThickness"
                    },
                },
                {
                    "label": "SpecimenType",
                    "value": "Fz 10x20",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/SpecimenType"
                    },
                },
                {
                    "label": "SampleIdentifier-2",
                    "value": "123456",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/SampleIdentifier-2"
                    },
                },
                {
                    "label": "ProjectNumber",
                    "value": "Projekt_1",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/ProjectNumber"
                    },
                },
                {
                    "label": "Tester",
                    "value": "Fe",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/Tester"
                    },
                },
                {
                    "label": "TestingRate",
                    "measurement_unit": {
                        "iri": "http://qudt.org/vocab/unit/MilliM-PER-SEC",
                        "label": "Millimetre per Second",
                        "namespace": "http://qudt.org/vocab/unit",
                        "symbol": "mm/s",
                    },
                    "value": 0.02,
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/TestingRate"
                    },
                },
                {
                    "label": "MachineData",
                    "value": "M_1",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/MachineData"
                    },
                },
                {
                    "label": "Temperature",
                    "measurement_unit": {
                        "iri": "http://qudt.org/vocab/unit/DEG_C",
                        "label": "degree Celsius",
                        "namespace": "http://qudt.org/vocab/unit",
                        "symbol": "°C",
                    },
                    "value": 25,
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/Temperature"
                    },
                },
                {
                    "label": "Material",
                    "value": "Werkstoff_1",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/Material"
                    },
                },
            ],
            "name": "General",
        },
    ],
}

columns = [
    "TestTime",
    "StandardForce",
    "Extension",
    "PercentageElongation",
    "AbsoluteCrossheadTravel",
    "WidthChange",
]


normal_config = {"graph_identifier": "https://www.example.org"}
bad_config = {"graph_identifier": "https://www.example.org", "foorbar": 123}


def test_xlsx_parser_no_match_in_metadata_from_mapping() -> None:
    from rdflib import Graph

    from data2rdf.models import PropertyGraph, QuantityGraph
    from data2rdf.parsers import ExcelParser
    from data2rdf.warnings import MappingMissmatchWarning

    with pytest.warns(
        MappingMissmatchWarning, match="Concept with key"
    ) as warnings:
        parser = ExcelParser(
            raw_data=raw_data,
            mapping=os.path.join(
                mapping_folder, "bad_metadata_tensile_test_mapping.json"
            ),
            parser_args={"unit_from_macro": True, "dropna": True},
        )
    missmatches = [
        warning
        for warning in warnings
        if warning.category == MappingMissmatchWarning
    ]
    assert len(missmatches) == 3

    expected_graph = Graph()
    expected_graph.parse(expected)

    assert len(parser.general_metadata) == 11
    for row in parser.general_metadata:
        assert isinstance(row, QuantityGraph) or isinstance(row, PropertyGraph)

    assert len(parser.dataframe_metadata) == 6
    for row in parser.dataframe_metadata:
        assert isinstance(row, QuantityGraph)

    assert len(parser.dataframe.columns) == 6
    assert sorted(list(parser.dataframe.columns)) == sorted(columns)
    for name, column in parser.dataframe.items():
        assert len(column) == 460


def test_xlsx_parser_no_match_in_timeseries_from_mapping() -> None:
    from rdflib import Graph

    from data2rdf.models import PropertyGraph, QuantityGraph
    from data2rdf.parsers import ExcelParser
    from data2rdf.warnings import MappingMissmatchWarning

    with pytest.warns(
        MappingMissmatchWarning, match="Concept with key"
    ) as warnings:
        parser = ExcelParser(
            raw_data=raw_data,
            mapping=os.path.join(
                mapping_folder, "bad_timeseries_tensile_test_mapping.json"
            ),
            parser_args={"unit_from_macro": True, "dropna": True},
        )

    missmatches = [
        warning
        for warning in warnings
        if warning.category == MappingMissmatchWarning
    ]
    assert len(missmatches) == 3

    expected_graph = Graph()
    expected_graph.parse(expected)

    assert len(parser.general_metadata) == 12
    for row in parser.general_metadata:
        assert isinstance(row, QuantityGraph) or isinstance(row, PropertyGraph)

    assert len(parser.dataframe_metadata) == 5
    for row in parser.dataframe_metadata:
        assert isinstance(row, QuantityGraph)

    assert len(parser.dataframe.columns) == 5
    for name, column in parser.dataframe.items():
        assert len(column) == 460

    assert remove_ids(parser.to_dict(schema=dsms_schema)) == sort_entries(
        metadata
    )
    assert sort_entries(parser.to_dict()) == as_non_dsms_schema(metadata)


@pytest.mark.parametrize("config", [normal_config, bad_config])
def test_csv_parser_config(config) -> None:
    from rdflib import Graph

    from data2rdf.parsers import ExcelParser
    from data2rdf.warnings import MappingMissmatchWarning

    with pytest.warns(
        MappingMissmatchWarning, match="Concept with key"
    ) as warnings:
        parser = ExcelParser(
            raw_data=raw_data,
            mapping=os.path.join(mapping_folder, "tensile_test_mapping.json"),
            parser_args={"unit_from_macro": True, "dropna": True},
            config=config,
        )

    missmatches = [
        warning
        for warning in warnings
        if warning.category == MappingMissmatchWarning
    ]
    assert len(missmatches) == 1

    expected_graph = Graph()
    expected_graph.parse(expected)

    assert parser.graph.isomorphic(expected_graph)
    assert str(parser.graph.identifier) == config["graph_identifier"]
    assert remove_ids(parser.to_dict(schema=dsms_schema)) == sort_entries(
        metadata
    )
    assert sort_entries(parser.to_dict()) == as_non_dsms_schema(metadata)
    assert sorted(list(parser.dataframe.columns)) == sorted(columns)


@pytest.mark.parametrize("extension", ["xlsx", "json", "csv", dict])
def test_parser_excel(extension) -> None:
    from rdflib import Graph

    from data2rdf.models import PropertyGraph, QuantityGraph
    from data2rdf.parsers import ExcelParser
    from data2rdf.warnings import MappingMissmatchWarning

    if isinstance(extension, str):
        mapping = os.path.join(
            mapping_folder, f"tensile_test_mapping.{extension}"
        )

    else:
        path = os.path.join(mapping_folder, "tensile_test_mapping.json")
        with open(path, encoding="utf-8") as file:
            mapping = json.load(file)

    with pytest.warns(
        MappingMissmatchWarning, match="Concept with key `Bemerkungen`"
    ) as warnings:
        parser = ExcelParser(
            raw_data=raw_data,
            mapping=mapping,
            parser_args={"unit_from_macro": True, "dropna": True},
        )

    missmatches = [
        warning
        for warning in warnings
        if warning.category == MappingMissmatchWarning
    ]
    assert len(missmatches) == 1

    assert len(parser.general_metadata) == 12
    for row in parser.general_metadata:
        assert isinstance(row, QuantityGraph) or isinstance(row, PropertyGraph)

    assert len(parser.dataframe_metadata) == 6
    for row in parser.dataframe_metadata:
        assert isinstance(row, QuantityGraph)

    assert len(parser.dataframe.columns) == 6
    assert sorted(list(parser.dataframe.columns)) == sorted(columns)
    for name, column in parser.dataframe.items():
        assert len(column) == 460

    expected_graph = Graph()
    expected_graph.parse(expected)

    assert parser.graph.isomorphic(expected_graph)

    assert remove_ids(parser.to_dict(schema=dsms_schema)) == sort_entries(
        metadata
    )
    assert sort_entries(parser.to_dict()) == as_non_dsms_schema(metadata)


@pytest.mark.parametrize("input_kind", ["path", "content"])
def test_parser_excel_inputs(input_kind) -> None:
    from rdflib import Graph

    from data2rdf.models import PropertyGraph, QuantityGraph
    from data2rdf.parsers import ExcelParser
    from data2rdf.warnings import MappingMissmatchWarning

    if input_kind == "path":
        input_obj = raw_data
    elif input_kind == "content":
        with open(raw_data, "rb") as file:
            input_obj = file.read()

    with pytest.warns(
        MappingMissmatchWarning, match="Concept with key `Bemerkungen`"
    ) as warnings:
        parser = ExcelParser(
            raw_data=input_obj,
            mapping=os.path.join(mapping_folder, "tensile_test_mapping.json"),
            parser_args={"unit_from_macro": True, "dropna": True},
        )
    missmatches = [
        warning
        for warning in warnings
        if warning.category == MappingMissmatchWarning
    ]
    assert len(missmatches) == 1

    assert len(parser.general_metadata) == 12
    for row in parser.general_metadata:
        assert isinstance(row, QuantityGraph) or isinstance(row, PropertyGraph)

    assert len(parser.dataframe_metadata) == 6
    for row in parser.dataframe_metadata:
        assert isinstance(row, QuantityGraph)

    assert len(parser.dataframe.columns) == 6
    assert sorted(list(parser.dataframe.columns)) == sorted(columns)
    for name, column in parser.dataframe.items():
        assert len(column) == 460

    expected_graph = Graph()
    expected_graph.parse(expected)

    assert parser.graph.isomorphic(expected_graph)

    assert remove_ids(parser.to_dict(schema=dsms_schema)) == sort_entries(
        metadata
    )
    assert sort_entries(parser.to_dict()) == as_non_dsms_schema(metadata)



"""Data2RDF utils"""
import tempfile
import warnings
from functools import lru_cache
from typing import Any, Dict, List, Optional

import requests
from rdflib import Graph

from data2rdf.warnings import QUDTMappingWarning


def _qudt_sparql(symbol: str) -> str:
    return f"""PREFIX qudt: <http://qudt.org/schema/qudt/>
    SELECT DISTINCT ?unit
        WHERE {{
            ?unit a qudt:Unit .
            {{
                ?unit qudt:symbol "{symbol}" .
            }}
            UNION
            {{
                ?unit qudt:ucumCode "{symbol}"^^qudt:UCUMcs .
            }}
        }}"""


@lru_cache
def _get_qudt_ontology(qudt_iri: str) -> requests.Response:
    response = requests.get(qudt_iri)
    if response.status_code != 200:
        raise RuntimeError(
            f"Could not download QUDT ontology. Please check URI: {qudt_iri}"
        )
    response.encoding = "utf-8"
    return response


def _to_tempfile(content) -> str:
    with tempfile.NamedTemporaryFile(
        mode="w", suffix=".ttl", delete=False, encoding="utf-8"
    ) as tmp:
        tmp.write(content)
    return tmp.name


@lru_cache
def _get_qudt_graph(qudt_iri: str) -> Graph:
    response = _get_qudt_ontology(qudt_iri)
    file = _to_tempfile(response.text)

    graph = Graph()
    graph.parse(file, encoding="utf-8")
    return graph


def _get_query_match(symbol: str, qudt_iri: str) -> List[str]:
    graph = _get_qudt_graph(qudt_iri)
    query = _qudt_sparql(symbol)
    return [str(row["unit"]) for row in graph.query(query)]


def _check_qudt_mapping(symbol: Optional[str]) -> Optional[str]:
    if symbol:
        match = _get_query_match(symbol)
        if len(match) == 0:
            warnings.warn(
                f"No QUDT Mapping found for unit with symbol `{symbol}`."
            )
            unit = {}
        elif len(match) > 1:
            warnings.warn(
                f"Multiple QUDT Mappings found for unit with symbol `{symbol}`."
            )
            unit = {
                "qudt:hasUnit": [
                    {"@value": uri, "@type": "xsd:anyURI"} for uri in match
                ]
            }
        else:
            unit = {
                "qudt:hasUnit": {"@value": match.pop(), "@type": "xsd:anyURI"}
            }
    else:
        unit = {}
    return unit


def _get_qudt_label_and_symbol(
    iri: str, qudt_iri: str, language: str
) -> Dict[str, Any]:
    graph = _get_qudt_graph(qudt_iri)
    gen_query = f"""PREFIX qudt: <http://qudt.org/schema/qudt/>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    SELECT DISTINCT ?label ?symbol
    WHERE {{
        <{iri}> qudt:symbol ?symbol .
        OPTIONAL {{
            <{iri}> rdfs:label ?label .
            FILTER (LANG(?label) = "{language}")
        }}
        OPTIONAL {{
            <{iri}> rdfs:label ?label_no_lang .
            FILTER (LANG(?label_no_lang) = "")
        }}
        BIND(COALESCE(?label, ?label_no_lang) AS ?label)
    }}"""
    match = [
        {"label": str(row["label"]), "symbol": str(row["symbol"])}
        for row in graph.query(gen_query)
    ]
    if len(match) == 0:
        warnings.warn(
            f"No QUDT label and symbol found for unit with iri `{iri}`.",
            QUDTMappingWarning,
        )
        unit = {}
    elif len(match) > 1:
        warnings.warn(
            f"Multiple QUDT symbols and labels found for unit with iri `{iri}`.",
            QUDTMappingWarning,
        )
        unit = match[0]
    else:
        unit = match.pop()
    return unit



"""Data2RDF"""

from .config import Config
from .parsers import Parser
from .pipelines import Data2RDF

from .models import (  # isort:skip
    ABoxBaseMapping,
    BasicConceptMapping,
    PropertyGraph,
    QuantityGraph,
)

__all__ = [
    "Data2RDF",
    "Config",
    "QuantityGraph",
    "PropertyGraph",
    "ABoxBaseMapping",
    "BasicConceptMapping",
    "Parser",
]
